from flask import Flask
from flask_restplus import Api
from werkzeug.utils import redirect

app = Flask(__name__)
api = Api(app)

# ensure slashes in URIs are matched as specified
# see https://stackoverflow.com/a/33285603/890242
app.url_map.strict_slashes = True
# use Flask json encoder to support datetime
app.config['RESTPLUS_JSON'] = {'cls': app.json_encoder}


@app.route('/docs')
def docs():
    return redirect("https://omegaml.github.io/omegaml/", code=302)



"""
omegaml write and read large CSV files from/to remote or local file paths
Supports s3, hdfs, http/s, sftp, scp, ssh

Usage:
    # read
    om.read_csv('protocol://path/test.csv', 'foobar')

    # write
    mdf = om.datasets.getl('tripdata')
    mdf.to_csv('protocol://path/test.csv')

    -- note path can be local, s3, http/s, scp, ftp, ssh (anything that smart_open supports,
       see https://github.com/RaRe-Technologies/smart_open

(c) 2019, 2020 omegaml.io by oneseven GmbH, Zurich, Switzerland
"""
import pandas as pd
from tqdm import tqdm

try:
    from smart_open import open as open_file
except:
    open_file = open


class IOToolsMDFMixin:
    def to_csv(mdf, csvfn, chunksize=10000, maxobs=None, apply=None, mode='w', open_kwargs=None, **kwargs):
        """
        write MDataframe to s3, hdfs, http/s, sftp, scp, ssh, write to om.datasets

        Usage:
            mdf = om.datasets.getl('name')
            mdf.to_csv('/path/filename.csv')
        """
        df_iter = mdf.iterchunks(chunksize=chunksize)
        _chunked_to_csv(df_iter, csvfn, mode, apply, open_kwargs=open_kwargs, **kwargs)


class IOToolsStoreMixin:
    def read_csv(self, csvfn, name, chunksize=10000, append=False, apply=None, mode='r',
                 open_kwargs=None, **kwargs):
        """
        read large files from s3, hdfs, http/s, sftp, scp, ssh, write to om.datasets

        Usage:
            om.datasets.read_csv('/path/filename.csv', 'dataset-name')

            Optionally define a process function that receives each dataframe chunk for processing:

            def process(df):
                # apply any processing to df
                return df

            om.datasets.read_csv(...., apply=process)
        """
        store = self
        open_kwargs = open_kwargs or {}
        with open_file(csvfn, mode=mode, **open_kwargs) as fin:
            it = pd.read_csv(fin, chunksize=chunksize, iterator=True, **kwargs)
            pbar = tqdm(it)
            try:
                for i, chunkdf in enumerate(pbar):
                    if apply:
                        result = apply(chunkdf)
                        chunkdf = chunkdf if result is None else result
                    store.put(chunkdf, name, append=(i > 0) or append)
            finally:
                pbar.close()
        return store.getl(name)

    def to_csv(self, name, csvfn, chunksize=10000, apply=None, mode='w', open_kwargs=None, **kwargs):
        """
        write an object that returns an iterable dataframe to s3, hdfs, http/s, sftp, scp, ssh

        Usage:
            om.datasets.write_csv(name, '/path/to/filename')
        """
        df_iter = self.get(name, chunksize=chunksize)
        _chunked_to_csv(df_iter, csvfn, mode, apply, open_kwargs=open_kwargs, **kwargs)


def _chunked_to_csv(df_iter, csvfn, mode, apply, open_kwargs=None, **kwargs):
    open_kwargs = open_kwargs or {}
    with open_file(csvfn, mode, **open_kwargs) as fout:
        for i, chunkdf in tqdm(enumerate(df_iter)):
            if apply:
                result = apply(chunkdf)
                chunkdf = chunkdf if result is None else result
            chunkdf.to_csv(fout, **kwargs)



class PythonPackageMixin(object):
    """
    Install and load scripts
    """
    def install(self, specs=None, keep=False):
        """
        install and load packages

        Args:
            specs (str, list): optional, package name or list of names,
                defaults to self.list()
            keep (bool): optional, will keep any packages installed in sys.path,
                defaults to False
        """
        specs = specs or self.list()
        if isinstance(specs, str):
            specs = specs.split(' ')
        for pkgname in specs:
            self.get(pkgname, keep=keep)



import pandas as pd

from omegaml.mdataframe import MDataFrame, MSeries


class UtilitiesMixin:
    """
    Add functionality to MDataFrame, MSeries
    """
    standard_quantiles = [.25, .5, .75]

    @property
    def dtypes(self):
        """
        Implement MDataFrame.dtypes to list all dtypes

        Returns:
            dtypes based on first row of data
        """
        if isinstance(self, MDataFrame):
            cursor = self.collection.find(limit=1)
            return self._get_dataframe_from_cursor(cursor)[self.columns].dtypes
        raise AttributeError('dtypes')

    @property
    def dtype(self):
        """
        Implement MSeries.dtype to get the dtype of a column

        Returns:
            dtype based on first row of data
        """
        if isinstance(self, MSeries):
            cursor = self.collection.find(limit=1)
            return self._get_dataframe_from_cursor(cursor)[self.name].dtype
        raise AttributeError('dtypes')

    def describe(self, quantiles=None):
        """
        implement MDataFrame.describe()

        Args:
            quantiles: a list of quantiles to compute, defaults to .25, .5, .75

        Returns:
            dataframe with quantiles
        """
        import numpy as np
        dtypes = self.dtypes
        # stats
        stats = ['mean', 'std', 'min', 'max']
        numcols = [col for col in dtypes.index
                   if np.issubdtype(dtypes[col], np.number)]
        specs = {col: stats for col in numcols}
        stats_df = self.apply(lambda v: v.agg(**specs)).value
        melted = stats_df.melt()
        melted['stat'] = (melted['variable']
                          .str.split('_')
                          .apply(lambda v: v[-1]))
        melted['variable'] = (melted['variable']
                              .str.split('_')
                              .apply(lambda v: '_'.join(v[:-1])))
        stats_df = melted.pivot_table(index='stat',
                                      columns='variable',
                                      values='value')
        # quantiles
        if quantiles:
            if not isinstance(quantiles, (tuple, list)):
                quantiles = self.standard_quantiles
            quants_df = self.quantile(quantiles).value
            stats_df = pd.concat([stats_df, quants_df], sort=False)
        return stats_df[numcols]

    def _amend_pipeline(self, pipeline):
        pipeline = super()._amend_pipeline(pipeline)
        if self.head_limit:
            pipeline.insert(0, {'$limit': self.head_limit})
        if self.skip_topn:
            pipeline.insert(0, {'$skip': self.skip_topn})
        return pipeline



def hello(**kwargs):
    return "hello from helloworld", kwargs


def run(om, **kwargs):
    """
    the script API execution entry point
    :return: result
    """
    import pandas as pd
    df = pd.DataFrame({
        'a': list(range(0, int(1e6 + 1))),
        'b': list(range(0, int(1e6 + 1)))
    })
    store = om.datasets
    store.put(df, 'mydata-xlarge', append=False, chunksize=50000)
    result = hello(**kwargs)
    return result



from contextlib import contextmanager

from celery import chain, group, chord


class CanvasTask:
    """
    support for canvas tasks

    See Also
        om.runtime.sequence
        om.runtime.parallel
        om.runtime.mapreduce
    """

    def __init__(self, canvasfn):
        self.sigs = []
        self.canvasfn = canvasfn
        self.runtime = None

    def add(self, task):
        self.sigs.append(task)

    def delay(self, *args, **kwargs):
        return self.apply_async(args=args, kwargs=kwargs)

    def apply_async(self, args=None, kwargs=None, **celery_kwargs):
        task = self.sigs[-1]
        task._apply_kwargs(kwargs, celery_kwargs)
        if self.canvasfn is chord:
            sig = task.task.signature(args=args, kwargs=kwargs, **celery_kwargs, immutable=False)
        else:
            # immutable means results are not passed on from task to task
            sig = task.task.signature(args=args, kwargs=kwargs, **celery_kwargs, immutable=True)
        self.sigs[-1] = sig
        return sig

    def run(self):
        if self.canvasfn is chord:
            result = self.canvasfn(self.sigs[:-1])(self.sigs[-1])
        else:
            result = self.canvasfn(*self.sigs).apply_async()
        return result


def make_canvased(canvasfn):
    @contextmanager
    def canvased(self):
        """
        context manager to support sequenced, parallel and mapreduce tasks

        Usage:
            # run tasks async, in sequence
            with om.runtime.sequence() as crt:
                crt.model('mymodel').fit(...)
                crt.model('mymodel').predict(...)
                result = crt.run()

            # run tasks async, in parallel
            with om.runtime.parallel() as crt:
                crt.model('mymodel').predict(...)
                crt.model('mymodel').predict(...)
                result = crt.run()

            # run tasks async, in parallel with a final step
            with om.runtime.mapreduce() as crt:
                # map tasks
                crt.model('mymodel').predict(...)
                crt.model('mymodel').predict(...)
                # reduce results - combined is a virtualobj function
                crt.model('combined').reduce(...)
                result = crt.run()

            # combined is a virtual obj function, e.g.
            @virtualobj
            def combined(data=None, **kwargs):
                # data is the list of results from each map step
                return data

            Note that the statements inside the context are
            executed in sequence, as any normal python code. However,
            the actual tasks are only executed on calling crt.run()

        Args:
            self: the runtime

        Returns:
            None
        """
        canvas = CanvasTask(canvasfn)
        _orig_task = self.task

        def canvas_task(*args, **kwargs):
            task = _orig_task(*args, **kwargs)
            canvas.add(task)
            return canvas

        self.task = canvas_task
        canvas.runtime = self
        canvas.runtime.run = canvas.run
        try:
            yield canvas.runtime
        finally:
            canvas.runtime.task = _orig_task
            canvas.runtime.run = None

    return canvased


canvas_chain = make_canvased(chain)
canvas_group = make_canvased(group)
canvas_chord = make_canvased(chord)



import six

from omegaml.backends.basemodel import BaseModelBackend
from omegaml.util import ensure_python_array


class KerasBackend(BaseModelBackend):
    KIND = 'keras.h5'

    @classmethod
    def supports(self, obj, name, **kwargs):
        from keras import Sequential, Model
        return isinstance(obj, (Sequential, Model))

    def _package_model(self, model, key, tmpfn):
        # https://www.tensorflow.org/api_docs/python/tf/keras/models/save_model
        # defaults to h5 since TF 2.x. We keep with h5 for simplicity for now
        import tensorflow as tf
        kwargs = dict(save_format='h5') if tf.__version__.startswith('2') else {}
        model.save(tmpfn, **kwargs)
        return tmpfn

    def _extract_model(self, infile, key, tmpfn):
        # override to implement model loading
        from keras.engine.saving import load_model
        with open(tmpfn, 'wb') as pkgf:
            pkgf.write(infile.read())
        return load_model(tmpfn)

    def fit(self, modelname, Xname, Yname=None, validation_data=None,
            pure_python=True, **kwargs):
        """
        Fit a model

        Args:
            modelname: the name of the model
            Xname: the name of the X dataset
            Yname: the name of the Y dataset
            pure_python: deprecated
            kwargs: any additional kwargs are passed on to model.fit()

        Returns:
            the meta data object of the model
        """
        meta = self.model_store.metadata(modelname)
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname) if Yname else None
        keras_kwargs = dict(kwargs)
        if validation_data:
            valX, valY = validation_data
            if isinstance(X, six.string_types):
                valX = self.data_store.get(valX)
            if isinstance(Y, six.string_types):
                valY = self.data_store.get(valY)
            keras_kwargs['validation_data'] = (valX, valY)
        history = model.fit(X, Y, **keras_kwargs)
        meta = self.model_store.put(model, modelname, attributes={
            'history': serializable_history(history)
        })
        return meta

    def predict(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        result = model.predict(X)
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result

    def score(
          self, modelname, Xname, Yname=None, rName=True, pure_python=True,
          **kwargs):
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname)
        result = model.evaluate(X, Y)
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result


def serializable_history(history):
    # ensure history object is JSON/BSON serializable by converting to default
    return {k: ensure_python_array(v, float)
            for k, v in six.iteritems(history.history)}



from tqdm import tqdm
import joblib

LokyBackend = joblib.parallel.BACKENDS['loky']


class OmegaRuntimeBackend(LokyBackend):
    """
    omega custom parallel backend to print progress

    TODO: extend for celery dispatching
    """
    def __init__(self, *args, **kwargs):
        self._tqdm = None
        self._job_count = kwargs.pop('n_tasks', None)
        self._verbose = kwargs.pop('verbose', True)
        import multiprocessing as mp
        # get LokyBackend to run in Celery, see LokyBackend.effective_n_jobs
        # TODO replace mp with billiard
        mp.current_process().daemon = False
        super().__init__(*args, **kwargs)

    def start_call(self):
        if self._verbose:
            self.tqdm = tqdm(total=self._job_count, unit='tasks')
        self._orig_print_progress = self.parallel.print_progress
        self.parallel.print_progress = self.update_progress

    def update_progress(self):
        try:
            self.tqdm.update(1) if self._verbose else None
        except:
            pass

    def stop_call(self):
        try:
            self.tqdm.close()
        except:
            pass

    def terminate(self):
        try:
            self.tqdm.close()
        except:
            pass
        finally:
            super().terminate()

#: register joblib parallel omegaml  backend
joblib.register_parallel_backend('omegaml', OmegaRuntimeBackend)



from __future__ import absolute_import

import warnings
from copy import deepcopy
from importlib import import_module

import logging
import os
import six
import sys
import tempfile
import uuid
from shutil import rmtree
from six import string_types

try:
    import urlparse
except:
    from urllib import parse as urlparse

# support pandas < 1.0
try:
    from pandas import json_normalize
except Exception as e:
    try:
        from pandas.io.json import json_normalize
    except:
        raise e
json_normalize = json_normalize

# reset global settings
__settings = None


def is_dataframe(obj):
    try:
        import pandas as pd
        return isinstance(obj, pd.DataFrame)
    except:
        return False


def is_series(obj):
    try:
        import pandas as pd
        return isinstance(obj, pd.Series) and not isinstance(obj, pd.DataFrame)
    except:
        return False


def is_estimator(obj):
    try:
        from sklearn.base import BaseEstimator
        from sklearn.pipeline import Pipeline
        return isinstance(obj, (BaseEstimator, Pipeline))
    except:
        False


def is_ndarray(obj):
    try:
        import numpy as np
        return isinstance(obj, np.ndarray)
    except:
        False


def is_spark_mllib(obj):
    """
    # unlike scikit learn obj is not the actual model, but a specification of
    # the model for the spark server to create. so obj is the name of the
    # python class, e.g. obj=pyspark.mllib.clustering.KMeans
    """
    if isinstance(obj, string_types):
        return 'pyspark.mllib' in obj
    return False


def settings(reload=False):
    """ wrapper to get omega settings from either django or omegaml.defaults """
    from omegaml import _base_config as omdefaults
    global __settings
    if not reload and __settings is not None:
        return __settings
    try:
        # see if we're running as a django app
        from django.contrib.auth.models import User
        from django.conf import settings as djsettings  # @UnresolvedImport
        try:
            getattr(djsettings, 'SECRET_KEY')
        except Exception as e:
            from warnings import warn
            warn("Using omegaml.defaults because Django was not initialized."
                 "Try importing omegaml within a method instead of at the "
                 "module level")
            raise
        else:
            defaults = djsettings
    except Exception as e:
        # django failed to initialize, use omega defaults
        defaults = omdefaults
    else:
        # get default omega settings into django settings if not set there
        from omegaml import _base_config as omdefaults
        for k in dir(omdefaults):
            if k.isupper() and not hasattr(defaults, k):
                setattr(defaults, k, getattr(omdefaults, k))
    __settings = DefaultsContext(defaults)
    omdefaults.load_framework_support(vars=__settings)
    omdefaults.load_user_extensions(vars=__settings)
    return __settings


def override_settings(**kwargs):
    """ test support """
    cfgvars = settings()
    for k, v in six.iteritems(kwargs):
        setattr(cfgvars, k, v)
    # -- OMEGA_CELERY_CONFIG updates
    celery_config = getattr(cfgvars, 'OMEGA_CELERY_CONFIG', {})
    for k in [k for k in kwargs.keys() if k.startswith('OMEGA_CELERY')]:
        celery_k = k.replace('OMEGA_', '')
        celery_config[celery_k] = kwargs[k]
    setattr(cfgvars, 'OMEGA_CELERY_CONFIG', celery_config)


def delete_database():
    """ test support """
    from omegaml.mongoshim import MongoClient

    mongo_url = settings().OMEGA_MONGO_URL
    parsed_url = urlparse.urlparse(mongo_url)
    database_name = parsed_url.path[1:]
    # authenticate via admin db
    # see https://stackoverflow.com/a/20554285
    c = MongoClient(mongo_url, authSource='admin')
    c.drop_database(database_name)


def make_tuple(arg):
    if not isinstance(arg, (list, tuple)):
        arg = (arg,)
    return tuple(arg)


def make_list(arg):
    if not isinstance(arg, (list)):
        arg = list(arg)
    return arg


def flatten_columns(col, sep='_'):
    # source http://stackoverflow.com/a/29437514
    if not type(col) is tuple:
        return col
    else:
        new_col = ''
        for leveli, level in enumerate(col):
            if not level == '':
                if not leveli == 0:
                    new_col += sep
                new_col += level
        return new_col


CLASS_CACHE = {}


def load_class(requested_class):
    """
    Check if requested_class is a string, if so attempt to load
    class from module, otherwise return requested_class as is
    """
    import importlib
    if isinstance(requested_class, six.string_types):
        if requested_class in CLASS_CACHE:
            return CLASS_CACHE.get(requested_class)
        module_name, class_name = requested_class.rsplit(".", 1)
        try:
            m = importlib.import_module(module_name)
            cls = getattr(m, class_name)
            CLASS_CACHE[requested_class] = cls
            return cls
        except Exception as e:
            logging.debug(
                'could not load module %s for class %s due to %s. sys.path=%s' % (
                    module_name, class_name, str(e), str(sys.path)))
            raise
    return requested_class


def get_rdd_from_df(df):
    """
    takes a pandas df and returns a spark RDD
    """
    from pyspark import SparkContext, SQLContext
    from pyspark.mllib.linalg import Vectors
    sc = SparkContext.getOrCreate()
    from warnings import warn
    warn(
        "get_rdd_from_df creates a spark context, it is recommended"
        " that you use SparkContext.getOrCreate() to prevent multiple context"
        " creation")
    sqlContext = SQLContext(sc)
    spark_df = sqlContext.createDataFrame(df)
    rdd = spark_df.rdd.map(lambda data: Vectors.dense(
        [float(x) for x in data]))
    return rdd


def get_labeledpoints(Xname, Yname):
    """
    returns a labeledpoint RDD from the datasets provided
    """
    import omegaml as om
    from pyspark.mllib.regression import LabeledPoint
    # import from datastore
    X = om.datasets.get(Xname)
    Y = om.datasets.get(Yname)
    result = Y.join(X)
    # create labeled point
    rdd = get_rdd_from_df(result)
    labeled_point = rdd.map(lambda x: LabeledPoint(float(x[0]), x[1:]))
    return labeled_point


def get_labeled_points_from_rdd(rdd):
    """
    returns a labeledpoint from the RDD provided
    """
    from pyspark.mllib.regression import LabeledPoint
    return rdd.map(lambda x: LabeledPoint(float(x[0]), x[1:]))


def unravel_index(df, row_count=0):
    """
    convert index columns into dataframe columns

    index columns are stored in the dataframe, named '_idx#<n>_<name>'
    where n is the sequence and name is the nth index name.

    Use restore_index_columns_order to get back the original index columns
    in sequence.

    :param df: the dataframe
    :param row_count: the row_count base, ensures subsequent stores
      get differnt row ids
    :return: the unravelled dataframe, meta
    """
    # remember original names
    idx_meta = {
        'names': df.index.names,
    }
    # convert index names so we can restore them later
    store_idxnames = ['_idx#{}_{}'.format(i, name or i)
                      for i, name in enumerate(idx_meta['names'])]
    df.index.names = store_idxnames
    unravelled_df, idx_meta = df.reset_index(), idx_meta
    # store row ids
    unravelled_df['_om#rowid'] = unravelled_df.index.values + row_count
    # restore index names on original dataframe
    df.index.names = idx_meta['names']
    return unravelled_df, idx_meta


def restore_index_columns_order(columns):
    """
    from an iterable of column names get the index columns in sequence order

    index columns are named '_idx#<n>_<name>' where n is the sequence
    of the original index column and name is the name
    """

    def get_index_order(col):
        if '_idx#' in col:
            n = col.split('_')[1].split('#')[1]
        else:
            n = 0
        return n

    index_cols = (col for col in columns if isinstance(col, six.string_types) and col.startswith('_idx'))
    index_cols = sorted(index_cols, key=get_index_order)
    return index_cols


def restore_index(df, idx_meta, rowid_sort=True):
    """
    restore index proper

    :param df: the dataframe
    :param idx_meta: index metadata
    :param rowid_sort: whether to sort by row id. defaults to True
           If your query is already sorted in some specific way,
           specify False to keep the sort order.
    """
    # -- establish row order proper
    if rowid_sort and '_om#rowid' in df:
        df.sort_values('_om#rowid', inplace=True)
        del df['_om#rowid']
    # -- get index columns
    index_cols = restore_index_columns_order(df.columns)
    # -- set index columns
    result = df.set_index(index_cols) if index_cols else df
    if index_cols:
        result.index.names = idx_meta.get('names', [None] * len(index_cols))
    return result


def jsonescape(s):
    return str(s).replace('.', '_')


def grouper(n, iterable):
    # https://stackoverflow.com/a/8998040
    import itertools
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


def cursor_to_dataframe(cursor, chunk_size=10000, parser=None):
    # a faster and less memory hungry variant of DataFrame.from_records
    # works by building a set of smaller dataframes to reduce memory
    # consumption. Note chunks are of size max. chunk_size.
    import pandas as pd
    frames = []
    if hasattr(cursor, 'count'):
        count = cursor.count()
        chunk_size = max(chunk_size, int(count * .1))
    else:
        # CommandCursors don't have .count, go as long as we can
        count = None
    if count is None or count > 0:
        for chunk in grouper(chunk_size, cursor):
            df = pd.DataFrame.from_records(chunk) if not parser else parser(r for r in chunk)
            frames.append(df)
        if frames:
            df = pd.concat(frames)
        else:
            df = pd.DataFrame()
    else:
        df = pd.DataFrame()
    return df


def ensure_index(coll, idx_specs, replace=False, **kwargs):
    """
    ensure a pymongo index specification exists on a given collection

    Checks if the index exists in regards to the fields in the specification.
    Only checks for field names, not sort order.

    Args:
        coll (pymongo.Collection): mongodb collection
        idx_specs (dict): specs as field => sort order

    Returns:
        None
    """
    idx_keys = list(dict(dict(v)['key']).keys() for v in coll.list_indexes())
    index_exists = any(all(k in keys for k in dict(idx_specs).keys()) for keys in idx_keys)
    idx_specs_SON = list(dict(idx_specs).items())
    created = False
    if index_exists and replace:
        coll.drop_index(idx_specs_SON, **kwargs)
        index_exists = False
    if not index_exists:
        coll.create_index(idx_specs_SON, **kwargs)
        created = True
    return created


def reshaped(data):
    """
    check if data is 1d and if so reshape to a column vector
    """
    import pandas as pd
    import numpy as np
    if isinstance(data, (pd.Series, pd.DataFrame)):
        if len(data.shape) == 1:
            data = data.values.reshape(-1, 1)
        else:
            if len(data.shape) == 2 and data.shape[1] == 1:
                data = data.values.reshape(-1, 1)
    elif isinstance(data, np.ndarray):
        if len(data.shape) == 1:
            data = data.reshape(-1, 1)
    elif isinstance(data, list):
        data = np.array(data)
        if len(data.shape) == 1:
            data = data.reshape(-1, 1)
    return data


def gsreshaped(data):
    """
    gridsearch reshape values according to GridSearchCV.fit

    see https://stackoverflow.com/a/49241326
    """
    import pandas as pd
    import numpy as np
    if isinstance(data, (pd.Series, pd.DataFrame)):
        if len(data.shape) == 2 and data.shape[1] == 1:
            data = data.values.reshape(-1)
    elif isinstance(data, np.ndarray):
        if len(data.shape) == 2 and data.shape[1] == 1:
            data = data.reshape(-1)
    return data


def convert_dtypes(df, dtypes):
    """
    get back original dtypes

    :param df: the dataframe to apply conversion to
    :param dtypes: the dict mapping column to dtype (use kind_meta['dtypes'])
    """
    # tz pattern used in convert_dtypes
    tzinfo_pattern = re.compile('datetime64\[ns, (.*)\]')
    for col, dtype in six.iteritems(dtypes):
        if dtype.startswith('datetime'):
            if not hasattr(df, 'dtypes'):
                continue
            try:
                match = tzinfo_pattern.match(dtype)
                if match:
                    tzname = match.groups()[0]
                    df[col] = df[col].dt.tz_localize('UTC').dt.tz_convert(tzname)
            except:
                # TODO ignore errors, issue warning
                pass

    return df


class PickableCollection(object):
    """
    A pickable pymongo.Collection

    This enables clean transmission of Collection instances to other processes, e.g.
    for multiprocessing tasks. It works by pickling the collection's connection data
    and restoring the database connection on unpickling.

    Usage:
        from multiprocessing import Pool

        def process(coll):
            coll.insert(...)

        p = Pool()
        p.map(process, range(1000),
    """

    def __init__(self, collection):
        super(PickableCollection, self).__setattr__('collection', collection)

    def __getattr__(self, k):
        return getattr(self.collection, k)

    def __setattr__(self, k, v):
        return setattr(self.collection, k, v)

    def __getitem__(self, k):
        return self.collection[k]

    def __setitem__(self, k, v):
        self.collection[k] = v

    def __getstate__(self):
        client = self.database.client
        host, port = list(client.nodes)[0]
        # options contains ssl settings
        options = self.database.client._MongoClient__options._options
        creds = self.database.client._MongoClient__all_credentials[self.database.name]
        creds_state = dict(creds._asdict())
        creds_state.pop('cache')
        creds_state['source'] = str(creds.source)
        return {
            'name': self.name,
            'database': self.database.name,
            'host': host,
            'port': port,
            'credentials': creds_state,
            'options': options,
        }

    def __setstate__(self, state):
        from omegaml.mongoshim import MongoClient
        url = 'mongodb://{username}:{password}@{host}:{port}/{database}'.format(**state, **state['credentials'])
        # ClientOptions calls it serverselectiontimeoutms, but stores seconds
        # MongoClient however, on recreating in unpickling, requires milliseconds
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html?highlight=serverSelectionTimeoutMS#pymongo.mongo_client.MongoClient
        # https://github.com/mongodb/mongo-python-driver/blob/7a539f227a9524b27ef469826ef9ee5bd4533773/pymongo/common.py
        # https://github.com/mongodb/mongo-python-driver/blob/master/pymongo/client_options.py#L157
        options = state['options']
        options['serverSelectionTimeoutMS'] = options.pop('serverselectiontimeoutms', 30) * 1000
        client = MongoClient(url, authSource=state['credentials']['source'], **options)
        db = client.get_database()
        collection = db[state['name']]
        super(PickableCollection, self).__setattr__('collection', collection)

    def __repr__(self):
        return 'PickableCollection({})'.format(repr(self.collection))


def extend_instance(obj, cls, *args, **kwargs):
    """Apply mixins to a class instance after creation"""
    # source https://stackoverflow.com/a/31075641
    from omegaml import load_class
    cls = load_class(cls)
    base_mro = obj.__class__.mro()
    if cls not in base_mro:
        base_cls = obj.__class__
        base_cls_name = 'Extended{}'.format(obj.__class__.__name__.split('.')[0])
        base_cls_name = base_mro[-2].__name__
        obj.__class__ = type(base_cls_name, (cls, base_cls), {})
    if hasattr(obj, '_init_mixin'):
        obj._init_mixin(*args, **kwargs)


def temp_filename(dir=None, ext='tmp'):
    """ generate a temporary file name """
    dir = dir or tempfile.mkdtemp()
    return os.path.join(dir, uuid.uuid4().hex + f'.{ext}')


def remove_temp_filename(fn, dir=True):
    dirname = os.path.dirname(fn)
    os.remove(fn)
    if dir:
        if dirname.startswith('/tmp/') and len(dirname.split('/')) > 1:
            rmtree(dirname)
        else:
            warnings.warn('will not remove directory {} as it is outside of /tmp'.format(fn))


def ensure_python_array(arr, dtype):
    import numpy as np
    return np.array(arr).astype(dtype)


def dict_update_if(condition, dict, other):
    try:
        condition()
    except:
        pass
    else:
        dict.update(other)


def module_available(modname):
    try:
        import_module(modname)
    except:
        return False
    return True


def tensorflow_available():
    return module_available('tensorflow')


def keras_available():
    return module_available('keras')


def calltrace(obj):
    """
    trace calls on an object

    Usage:
        .. code ::

            def __init__(self, *args, **kwargs):
                calltrace(self)

        This will print the method arguments and return values
        for every call. Exceptions will be caught and printed,
        then re-raised. Magic methods are not traced.

    Args:
        obj: the object to trace
    """

    def tracefn(el):
        def methodcall(*args, **kwargs):
            print("calling", el, args, kwargs)
            try:
                result = el(*args, **kwargs)
            except (BaseException, Exception) as e:
                print("=> ", str(e))
                raise
            except:
                print("=> unknown exception")
                raise
            else:
                print("=> ", result)
            return result

        return methodcall

    for item in dir(obj):
        if item.startswith('__'):
            continue
        el = getattr(obj, item)
        if type(el).__name__ == 'method':
            setattr(obj, item, tracefn(el))
    return obj


class DefaultsContext(object):
    """
    om.defaults as set for a particular Omega() instance

    Usage:
        defaults = DefaultsContext(source)

        # attribute access
        defaults.SOME_VARIABLE
        defaults['SOME_VARIABLE']
        defaults.get('SOME_VARIABLE', default=None)

        were source is a module or a settings object (any object with
        any number of .UPPERCASE_VARIABLE attributes). Source will be
        deep-copied to ensure DefaultsContext cannot be changed by external
        references.
    """

    def __init__(self, source):
        for k in dir(source):
            if k.isupper():
                value = getattr(source, k)
                setattr(self, k, deepcopy(value))

    def __iter__(self):
        for k in dir(self):
            if k.startswith('OMEGA') and k.isupper():
                yield k, getattr(self, k)

    def __getitem__(self, k):
        if k in dir(self):
            return getattr(self, k)
        raise KeyError(k)

    def __setitem__(self, k, v):
        setattr(self, k, v)

    def get(self, k, default=None):
        try:
            return self[k]
        except KeyError:
            return default

    def __repr__(self):
        return 'DefaultsContext({})'.format(dict(self))


def ensure_json_serializable(v):
    import numpy as np
    import pandas as  pd
    if isinstance(v, np.ndarray):
        return v.flatten().tolist()
    if isinstance(v, pd.Series):
        return ensure_json_serializable(v.to_dict())
    elif isinstance(v, dict):
        vv = {
            k: ensure_json_serializable(v)
            for k, v in six.iteritems(v)
        }
        return vv
    return v


def mkdirs(path):
    """ safe os.makedirs for python 2 & 3
    """
    if not os.path.exists(path):
        os.makedirs(path)


def base_loader(_base_config):
    try:
        from omegaee import omega as _omega
        from omegaee import eedefaults as _base_config_ee
    except Exception as e:
        from omegaml import omega as _omega
    else:
        _base_config.update_from_obj(_base_config_ee, attrs=_base_config)
    settings(reload=True)
    return _omega


from io import StringIO

from contextlib import contextmanager
import re


def markup(file_or_str, parsers=None, direct=True, on_error='warn', default=None, msg='could not read {}',
           **kwargs):
    """
    a safe markup file reader, accepts json and yaml, returns a dict or a default
    Usage:
        file_or_str = filename|file-like|markup-str
        # try, return None if not readable, will issue a warning in the log
        data = markup(file_or_str)
        # try, return some other default, will issue a warning in the log
        data = markup(file_or_str, default={})
        # try and fail
        data = markup(file_or_str, on_error='fail')
    Args:
        file_or_str (None, str, file-like): any file-like, can be
           any object that the parsers accept
        parsers (list): the list of parsers, defaults to json.load, yaml.safe_load,
           json.loads
        direct (bool): if True returns the result, else returns markup (self). then use
           .read() to actually read the contents
        on_error (str): 'fail' raises a ValueError in case of error, 'warn' outputs a warning to the log,
           and returns the default, 'silent' returns the default. Defaults to warn
        default (obj): return the obj if the input is None or in case of on_error=warn or silent
        **kwargs (dict): any kwargs passed on to read(), any entry that matches a parser
            function's module name will be passed on to the parser
    Returns:
        data parsed or default
        markups.exceptions contains list of exceptions raised, if any
    """
    # source: https://gist.github.com/miraculixx/900a28a94c375b7259b1f711b93417d3
    import json
    import yaml
    import logging

    parsers = parsers or (json.load, yaml.safe_load, json.loads)
    # path-like regex
    # - \/?                  leading /, optional
    # - (?P<path>\w+/?)*     any path-part followed by /, repeated 0 - n times
    # - (?P<ext>(\w*\.?\w*)+ any file.ext, at least once
    pathlike = lambda s: re.match(r"^\/?(?P<path>\w+/?)*(?P<ext>(\w*\.?\w*))$", s)

    @contextmanager
    def fopen(filein, *args, **kwargs):
        # https://stackoverflow.com/a/55032634/890242
        if isinstance(filein, str) and pathlike(filein):  # filename
            with open(filein, *args, **kwargs) as f:
                yield f
        elif isinstance(filein, str):  # some other string, make a file-like
            yield StringIO(filein)
        else:
            # file-like object
            yield filein

    throw = lambda ex: (_ for _ in ()).throw(ex)
    exceptions = []

    def read(**kwargs):
        if file_or_str is None:
            return default
        for fn in parsers:
            try:
                with fopen(file_or_str) as fin:
                    if hasattr(fin, 'seek'):
                        fin.seek(0)
                    data = fn(fin, **kwargs.get(fn.__module__, {}))
            except Exception as e:
                exceptions.append(e)
            else:
                return data
        # nothing worked so far
        actions = {
            'fail': lambda: throw(ValueError("Reading {} caused exceptions {}".format(file_or_str, exceptions))),
            'warn': lambda: logging.warning(msg.format(file_or_str)) or default,
            'silent': lambda: default,
        }
        return actions[on_error]()

    markup.read = read
    markup.exceptions = exceptions
    return markup.read(**kwargs) if direct else markup


def raises(fn, wanted_ex):
    try:
        fn()
    except Exception as e:
        assert isinstance(e, wanted_ex), "expected {}, raised {} instead".format(wanted_ex, e)
    else:
        raise ValueError("did not raise {}".format(wanted_ex))
    return True


def dict_merge(destination, source, delete_on='__delete__', subset=None):
    """
    Merge two dictionaries, including sub dicts

    Args:
        destination (dict): the dictionary to merge into
        source (dict): the dictionary to merge from
        delete_on (obj): for each entry in source, its value is
            compared to match delete_on, if it does the key will
            be deleted in the destination dict. Defaults to '__delete__'
        subset (callable): optional, only merge item
            if subset(key, value) is True

    See Also:
        https://stackoverflow.com/a/20666342/890242
    """
    dict_merge.DELETE = delete_on
    for key, value in source.items():
        if callable(subset) and not subset(key, value):
            continue
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            dict_merge(node, value, delete_on=delete_on)
        else:
            if value == dict_merge.DELETE and key in destination:
                del destination[key]
            else:
                destination[key] = value


def ensure_base_collection(collection):
    """ get base from pymongo.Collection subclass instance """
    from pymongo.collection import Collection
    is_real_collection = isinstance(collection, Collection)
    while not is_real_collection:
        collection = collection.collection
        is_real_collection = isinstance(collection, Collection)
    return collection



import os

import unittest

from omegaml import Omega
from omegaml.documents import Metadata
from omegaml.util import settings as omegaml_settings


class JobTasksTests(unittest.TestCase):
    def setUp(self):
        super().setUp()
        # ensure that sub tasks run in test mode
        os.environ['OMEGA_TEST_MODE'] = '1'
        # ensure that sub tasks don't use authenticated runtime
        os.environ['OMEGA_ALLOW_TASK_DEFAULT_AUTH'] = 'yes'

        for omx in (self.om, self.om['bucket']):
            for fn in omx.jobs.list():
                omx.jobs.drop(fn)

    @property
    def om(self):
        om = Omega(defaults=omegaml_settings(reload=True))
        return om

    @property
    def fs(self):
        om = self.om
        defaults = omegaml_settings()
        fs = om.jobs.get_fs(defaults.OMEGA_NOTEBOOK_COLLECTION)
        return fs

    def test_map(self):
        """ test runtime.job.map() works ok """
        om = self.om
        meta = om.jobs.create("print('hello')", 'main')
        job = om.runtime.job('main')
        tasks = job.map(range(2))
        self.assertEqual(len(tasks), 2)
        self.assertTrue(all(isinstance(t, Metadata) for t in tasks))
        status = job.status().set_index('name')
        for t in tasks:
            self.assertIn('task_id', t.attributes)
            task_name = t.attributes['job']['task_name']
            result = t.attributes['job_runs'][-1]
            self.assertEqual(result['status'], 'OK')
            self.assertEqual(status.loc[t.name, 'run_status'], 'OK')

    def test_map_fail(self):
        """ test runtime.job.map() works with erronous jobs """
        om = self.om
        # introduce some error
        meta = om.jobs.create("raise ValueError()", 'main')
        job = om.runtime.job('main')
        tasks = job.map(range(2))
        self.assertEqual(len(tasks), 2)
        self.assertTrue(all(isinstance(t, Metadata) for t in tasks))
        status = job.status().set_index('name')
        for t in tasks:
            self.assertIn('task_id', t.attributes)
            task_name = t.attributes['job']['task_name']
            result = t.attributes['job_runs'][-1]
            self.assertEqual(result['status'], 'ERROR')
            self.assertEqual(status.loc[t.name, 'run_status'], 'ERROR')
        # fix error and restart
        meta = om.jobs.create("print('hello')", 'main')
        tasks = job.restart(reset=True)
        self.assertEqual(len(tasks), 2)
        self.assertTrue(all(isinstance(t, Metadata) for t in tasks))
        status = job.status().set_index('name')
        for t in tasks:
            self.assertIn('task_id', t.attributes)
            task_name = t.attributes['job']['task_name']
            result = t.attributes['job_runs'][-1]
            self.assertEqual(result['status'], 'ERROR')
            self.assertEqual(status.loc[t.name, 'run_status'], 'ERROR')

    def test_map_groups(self):
        """ test runtime.job.map() works with multiple groups """
        import os

        om = self.om
        meta = om.jobs.create("print('hello')", 'main')
        job = om.runtime.job('main')

        def check(tasks):
            self.assertEqual(len(tasks), 2)
            self.assertTrue(all(isinstance(t, Metadata) for t in tasks))
            status = job.status().set_index('name')
            for t in tasks:
                self.assertIn('task_id', t.attributes)
                task_name = t.attributes['job']['task_name']
                result = t.attributes['job_runs'][-1]
                self.assertEqual(result['status'], 'OK')
                self.assertEqual(status.loc[t.name, 'run_status'], 'OK')
            return tasks

        group1_tasks = job.map(range(2))
        group2_tasks = job.map(range(2))
        check(group1_tasks)
        check(group2_tasks)

        all_groups = job.status('*')
        self.assertEqual(len(all_groups), 4)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

import pandas as pd
from pandas.util.testing import assert_frame_equal, assert_series_equal

from omegaml import Omega
from omegaml.tests.util import OmegaTestMixin


class ParallelMixinTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()

    def test_parallel_process_persist(self):
        """
        parallel processing of mdf, persisting to a dataset
        """
        om = self.om
        large = pd.DataFrame({
            'x': range(1000)
        })

        def myfunc(df):
            df['y'] = df['x'] * 2

        om.datasets.put(large, 'largedf', append=False)
        mdf = om.datasets.getl('largedf')
        mdf.transform(myfunc).persist('largedf_transformed', om.datasets)
        self.assertIn('largedf_transformed', om.datasets.list())
        dfx = om.datasets.get('largedf_transformed')
        large['y'] = large['x'] * 2
        assert_frame_equal(dfx, large)

    def test_parallel_process_direct(self):
        """
        parallel processing of mdf, persisting to a dataset
        """

        om = self.om
        large = pd.DataFrame({
            'x': range(1000)
        })

        def myfunc(df):
            df['y'] = df['x'] * 2

        om.datasets.put(large, 'largedf', append=False)
        mdf = om.datasets.getl('largedf')
        dfx = mdf.transform(myfunc).value
        large['y'] = large['x'] * 2
        assert_frame_equal(dfx, large)

    def test_parallel_worker_resolv(self):
        """
        test worker resolves mdf, func receives native df
        """
        om = self.om
        large = pd.DataFrame({
            'x': range(1000)
        })

        def myfunc(df):
            df['y'] = df['x'] * 2

        om.datasets.put(large, 'largedf', append=False)
        mdf = om.datasets.getl('largedf')
        mdf.transform(myfunc).persist('largedf_transformed', om.datasets)
        self.assertIn('largedf_transformed', om.datasets.list())
        dfx = om.datasets.get('largedf_transformed')
        large['y'] = large['x'] * 2
        assert_frame_equal(dfx, large)

    def test_parallel_func_resolve(self):
        """
        test process function resolves mdf, return df
        """
        om = self.om
        large = pd.DataFrame({
            'x': range(1000)
        })

        def myfunc(mdf):
            df = mdf.value
            df['y'] = df['x'] * 2
            return df

        om.datasets.put(large, 'largedf', append=False)
        mdf = om.datasets.getl('largedf')
        mdf.transform(myfunc, resolve='func').persist('largedf_transformed', om.datasets)
        self.assertIn('largedf_transformed', om.datasets.list())
        dfx = om.datasets.get('largedf_transformed')
        large['y'] = large['x'] * 2
        assert_frame_equal(dfx, large)

    def test_parallel_custom_chunk(self):
        """
        test process function resolves mdf, return df
        """
        om = self.om
        large = pd.DataFrame({
            'x': range(100)
        })

        def chunker(mdf, chunksize, maxobs):
            for i in range(0, len(mdf), chunksize):
                yield mdf.skip(i).head(chunksize)

        def myfunc(df):
            df['y'] = df['x'] * 2

        om.datasets.put(large, 'largedf', append=False)
        mdf = om.datasets.getl('largedf')
        mdf.transform(myfunc, chunkfn=chunker, chunksize=5, n_jobs=1).persist('largedf_transformed', om.datasets)
        self.assertIn('largedf_transformed', om.datasets.list())
        # parallel insert may result in different row order, get back in sort order
        dfx = om.datasets.getl('largedf_transformed').sort('x').value
        large['y'] = large['x'] * 2
        self.assertEqual(len(dfx), len(large))
        assert_frame_equal(dfx.reset_index(), large.reset_index())




import os
from pprint import pprint
from time import sleep

from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class RuntimeCommandBase(CommandBase):
    """
    Usage:
      om runtime model <name> <model-action> [<X>] [<Y>] [--result=<output-name>] [--param=<kw=value>]... [--async] [options]
      om runtime script <name> [<script-action>] [<kw=value>...] [--async] [options]
      om runtime job <name> [<job-action>] [<args...>] [--async] [options]
      om runtime result <taskid> [options]
      om runtime ping [options]
      om runtime env <action> [<package>] [--file <requirements.txt>] [--require <label>] [--every] [options]
      om runtime log [-f] [options]
      om runtime status [workers|labels|stats] [options]
      om runtime restart app <name> [options]
      om runtime celery [<celery-command>...] [--worker=<worker>] [--queue=<queue>] [--celery-help] [--flags <celery-flags>...] [options]

    Options:
      --async           don't wait for results, will print taskid
      -f                tail log
      --require=VALUE   worker label
      --flags=VALUE     celery flags, list as "--flag VALUE"
      --worker=VALUE    celery worker
      --queue=VALUE     celery queue
      --celery-help     show celery help
      --file=VALUE      path/to/requirements.txt
      --local           if specified the task will run locally. Use this for testing
      --every           if specified runs task on all workers

    Description:
      model, job and script commands
      ------------------------------

      <model-action> can be any valid model action like fit, predict, score,
      transform, decision_function etc.

      <script-action> defaults to run
      <job-action> defaults to run

      Examples:
        om runtime model <name> fit <X> <Y>
        om runtime model <name> predict <X>
        om runtime job <name>
        om runtime script <name>
        om runtime script <name> run myparam="value"

      running asynchronously
      ----------------------

      model, job, script commands accept the --async paramter. This will submit
      the a task and return the task id. To wait for and get the result run use
      the result command

      Examples:
            om runtime model <name> fit <X> <Y> --async
            => <task id>
            om runtime result <task id>
            => result of the task

      restart app
      -----------

      This will restart the app on omegaml apphub. Requires a login to omegaml cloud.


      status
      ------

      Prints workers, labels, list of active tasks per worker, count of tasks

      Examples:
        om runtime status             # defaults to workers
        om runtime status workers
        om runtime status labels
        om runtime status stats

      celery commands
      ---------------

      This is the same as calling celery -A omegaml.celeryapp <commands>. Command
      commands include:

      inspect active         show currently running tasks
      inspect active_queues  show active queues for each worker
      inspect stats          show stats of each worker, including pool size (processes)
      inspect ping           confirm that worker is connected

      control pool_shrink N  shrink worker pool by N, specify 99 to remove all
      control pool_grow N    grow worker poool by N

      Examples:
            om runtime celery inspect active
            om runtime celery control pool_grow N


      env commands
      ------------

      This talks to an omegaml worker's pip environment

      a) install a specific package

         env install <package>    install the specified package, use name==version pip syntax for specific versions
         env uninstall <package>  uninstall the specified package

      b) use a requirements file

         env install --file requirements.txt
         env uninstall --file requirements.txt

      c) list currently installed packages

         env freez

      d) install on all or a specific worker

         env install --require gpu package
         env install --every package

         By default the installation runs on the default worker only. If there are multiple nodes where you
         want to install the package(s) worker nodes, be sure to specify --every

      Examples:
            om runtime env install pandas
            om runtime env uninstall pandas
            om runtime env install --file requirements.txt
            om runtime env install --file gpu-requirements.txt --require gpu
            om runtime env install --file requirements.txt --every
    """
    command = 'runtime'

    def ping(self):
        om = get_omega(self.args)
        label = self.args.get('--require')
        self.logger.info(om.runtime.require(label).ping())

    def _ensure_valid_XY(self, value):
        if value is not None:
            if value.isnumeric():
                return [eval(value)]
            if value[0] == '[' and value[-1] == ']':
                return eval(value)
        return value

    def model(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        action = self.args.get('<model-action>')
        is_async = self.args.get('--async')
        kwargs_lst = self.args.get('--param')
        output = self.args.get('--result')
        label = self.args.get('--require')
        X = self._ensure_valid_XY(self.args.get('<X>'))
        Y = self._ensure_valid_XY(self.args.get('<Y>'))
        # parse the list of kw=value values
        # e.g. key1=val1 key2=val2 => kwargs_lst = ['key1=val1', 'key2=val2']
        #   => kw_dct = { 'key1': eval('val1'), 'key2': eval('val2') }
        kv_dct = {}
        for kv in kwargs_lst:
            k, v = kv.split('=', 1)
            kv_dct[k] = eval(v)
        kwargs = {}
        if action in ('predict', 'predict_proba',
                      'decision_function', 'transform'):
            # actions that take rName, but no Y
            kwargs['rName'] = output
        else:
            # actions that take Y, but no rName
            kwargs['Yname'] = Y
        if action == 'gridsearch':
            kwargs['parameters'] = kv_dct
        rt_model = om.runtime.require(label).model(name)
        meth = getattr(rt_model, action, None)
        if meth is not None:
            result = meth(X, **kwargs)
            if not is_async:
                self.logger.info(result.get())
            else:
                self.logger.info(result)
            return
        raise ValueError('{action} is not applicable to {name}'.format(**locals()))

    def script(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        is_async = self.args.get('--async')
        kwargs = self.parse_kwargs('<kw=value>')
        label = self.args.get('--require')
        result = om.runtime.require(label).script(name).run(**kwargs)
        if not is_async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def job(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        is_async = self.args.get('--async')
        label = self.args.get('--require')
        result = om.runtime.require(label).job(name).run()
        if not is_async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def result(self):
        from celery.result import AsyncResult

        om = get_omega(self.args)
        task_id = self.args.get('<taskid>')
        result = AsyncResult(task_id, app=om.runtime.celeryapp).get()
        self.logger.info(result)

    def log(self):
        import pandas as pd
        tail = self.args.get('-f')
        om = get_omega(self.args)
        if not tail:
            df = om.logger.dataset.get()
            with pd.option_context('display.max_rows', None,
                                   'display.max_columns', None,
                                   'display.max_colwidth', -1):
                print(df[['text']])
        else:
            om.logger.dataset.tail()

    def celery(self):
        om = get_omega(self.args)
        celery_cmds = ['celery']
        # convert omega terms into celery terms
        celery_opts = (
            # omega term, celery term, value|flag
            ('--worker', '--destination', 'value'),
            ('--queue', '--queue', 'value'),
            ('--celery-help', '--help', 'flag'),
        )
        for opt, celery_opt, kind in celery_opts:
            if self.args.get(opt):
                celery_cmds += [celery_opt]
                if kind == 'value':
                    celery_cmds += [self.args.get(opt)]
        celery_cmds += self.args.get('<celery-command>')
        celery_cmds += self.args.get('--flags')
        om.runtime.celeryapp.start(celery_cmds)

    def env(self):
        om = get_omega(self.args)
        action = self.args.get('<action>')
        package = self.args.get('<package>')
        reqfile = self.args.get('--file')
        every = self.args.get('--every')
        require = self.args.get('--require') or ''
        if reqfile:
            with open(reqfile, 'rb') as fin:
                om.scripts.put(fin, '.system/requirements.txt')
        if not om.scripts.exists('.system/envinstall', hidden=True):
            import omegaml as om_module
            envinstall_path = os.path.join(os.path.dirname(om_module.__file__), 'runtimes', 'envinstall')
            om.scripts.put(f'pkg://{envinstall_path}', '.system/envinstall')
        if every:
            labels = om.runtime.enable_hostqueues()
        else:
            labels = require.split(',')
        results = []
        for label in labels:
            result = (om.runtime.require(label)
                      .script('.system/envinstall')
                      .run(action=action, package=package, file=reqfile,
                           __format='python'))
            results.append((label, result))
        all_results = om.runtime.celeryapp.ResultSet([r[1] for r in results])
        from tqdm import tqdm
        with tqdm() as progress:
            while all_results.waiting():
                progress.update(1)
                sleep(1)
            all_results.get()
        for label, result in results:
            if label:
                print(f'** result of worker require={label}:')
            data = result.get()  # resolve AsyncResult => dict
            print(str(data.get('result', data)))  # get actual result object, pip stdout

    def status(self):
        om = get_omega(self.args)
        labels = self.args.get('labels')
        stats = self.args.get('stats')
        workers = not (labels or stats)
        if workers:
            pprint(om.runtime.workers())
        elif labels:
            queues = om.runtime.queues()
            pprint({worker: [q.get('name') for q in details
                             if not q.get('name').startswith('amq')]
                    for worker, details in queues.items()})
        elif stats:
            stats = om.runtime.stats()
            pprint({worker: {
                'size': details['pool']['max-concurrency'],
                'tasks': {
                    task: count for task, count in details['total'].items()
                }
            } for worker, details in stats.items()})

    def restart(self):
        import requests
        om = get_omega(self.args, require_config=True)
        name = self.args.get('<name>')
        user = om.runtime.auth.userid
        auth = requests.auth.HTTPBasicAuth(user, om.runtime.auth.apikey)
        url = om.defaults.OMEGA_RESTAPI_URL
        stop = requests.get(f'{url}/apps/api/stop/{user}/{name}'.format(om.runtime.auth.userid),
                            auth=auth)
        start = requests.get(f'{url}/apps/api/start/{user}/{name}'.format(om.runtime.auth.userid),
                             auth=auth)
        self.logger.info(f'stop: {stop} start: {start}')
